import importlib 
import types
import inspect
import logging

from django.core.exceptions import ImproperlyConfigured   
from django.contrib.auth.models import User
from django.conf import settings 
from django.db import models        

from appconf import AppConf         

logger = logging.getLogger(__name__)

def check_achievement_class(cls):
    return [attribute for attribute in ['name', 'key', 'description', 'bonus', 'evaluate'] if not hasattr(cls, attribute)]      

def load_classes(classes=settings.ACHIEVEMENT_CLASSES, *args, **kwargs):  
    # if we're called during south migration ignore every app 
    # until we get notified that the achievements app 
    # was properly loaded
    current_app = kwargs.get('app', None)
    # this case is when we're using south migrations
    if type(current_app) is str and current_app != 'achievements':
        return                                              
    # otherwise when we use plain syncdb :
    if type(current_app) is types.ModuleType and current_app.__name__ != 'achievements.models':
        return
    from achievements.engine import engine
    for achievement_module in classes:
        try:
            module = importlib.import_module(achievement_module)
            clses = [cls for name, cls in inspect.getmembers(module) if inspect.isclass(cls) and name.endswith('Achievement')]
            for cl in clses:
                errors = check_achievement_class(cl)
                if errors:              
                    message = "Achievement class '%s' in '%s' has missing attributes : %s" %(cl.__name__, module.__name__, ",".join(errors))    
                    logger.error(message)
                    raise ImproperlyConfigured(message)
                else:
                    logger.info("Registering achievement class %s..." % (cl))
                    engine.register_achievement(cl)
        except ImproperlyConfigured:
            raise
        except Exception, exc:  
            logger.error("Exception catched while trying to register achievements class %s " % exc)
            raise ImproperlyConfigured("ACHIEVEMENT_CLASSES attribute must be set properly for them to be loaded into the engine : %s" % exc)   
    
                     

class Achievement(models.Model):
    """ These objects are what people are earning when contributing """  
    name = models.CharField(max_length=75)
    key = models.CharField(max_length=75, unique=True)
    description = models.TextField(null=True, blank=True)
    bonus = models.IntegerField(default=0)
    callback = models.TextField()  
    def __unicode__(self):
        return "Achievement(%s, %s)"% (self.name, self.bonus)  

class UserAchievement(models.Model):
    user = models.ForeignKey(User)
    achievement = models.ForeignKey(Achievement)     
    registered_at = models.DateTimeField(auto_now_add=True)  
                                
class AchievementEngineConf(AppConf): 
    """ Configuration class used by Django AppConf to ease the setup"""
    USE_CELERY = False
    CLASSES = []     
    
    class Meta:
        prefix = 'achievement'

    def configure_classes(self, value):
        pass  

# connect to the end of the syncdb command signal to reload achievements at that time.
if 'south' in settings.INSTALLED_APPS:
    from south.signals import post_migrate
    post_migrate.connect(load_classes)
else:
    from django.db.models.signals import post_syncdb
    post_syncdb.connect(load_classes)



import importlib    
import logging
                                     
from django.db.models.aggregates import Sum
from achievements.models import Achievement, UserAchievement         
from achievements.signals import achievement_registered, achievement_unlocked

logger = logging.getLogger(__name__)

def get_user_score(user):      
    """ Compute the score of a given user taking into account their Achievement's bonuses"""
    return UserAchievement.objects.filter(user=user).aggregate(score=Sum('achievement__bonus'))['score']     

def check_achievement_plain(sender, user, key, *args, **kwargs): 
    logger.debug("Check achievement called by %s" % sender)
    obj = Achievement.objects.get(key=key)
    if evaluate_achievement_callback(user, obj, *args, **kwargs):
        (user_ach, created) = UserAchievement.objects.get_or_create(achievement=obj, user=user)
        if created:     
            logger.info("Achievement %s unlocked for user %s " % (obj, user))
            achievement_unlocked.send(sender=sender, user=user, achievement=obj, *args, **kwargs)
                                        
def evaluate_achievement_callback(user, obj, *args, **kwargs):
    """ 
    Evaluate callback of the achievement in order to determine if achievement has been
    unlocked by user.
    """            
    achievement = get_callback_object(obj.callback)
    # call the evaluate fonction :
    return achievement().evaluate(user, *args, **kwargs)
                                                   
def construct_callback(obj):         
    logger.debug("Construction callback from %s : %s.%s " % (obj, obj.__module__, obj.__name__))
    return "%s.%s" % (obj.__module__, obj.__name__)

def get_callback_object(ref):
    module = ".".join(ref.split(".")[:-1])
    class_name = ref.split(".")[-1]    
    m = importlib.import_module(module)
    return eval('m.%s' % class_name)



from django.conf import settings                                                
from achievements.models import Achievement, UserAchievement                    
from achievements.utils import construct_callback, check_achievement_plain
from achievements.signals import achievement_registered, achievement_unlocked
            
class AchievementEngine(object):
    def register_achievement(self, cls):  
        """ Register a new achievement into the engine and database"""    
        (obj, created) = Achievement.objects.get_or_create(key=cls.key,  defaults={
            'name' : cls.name,
            'description' : cls.description, 
            'bonus' : cls.bonus, 
            'callback': construct_callback(cls)})
        if not created:
            # update the object if key didn't change :
            obj.name = cls.name
            obj.description = cls.description
            obj.bonus = cls.bonus
            obj.callback = construct_callback(cls) 
            obj.save()
        achievement_registered.send(sender=self, achievement_class=cls)      
                   
    def check_achievement(self, user, key, *args, **kwargs):
        """ 
        Check synchronously or asynchronously if according to a specific context 
        an achievement has been unlocked
        """
        if user and user.is_authenticated(): 
            if settings.ACHIEVEMENT_USE_CELERY: 
                # do not try to import if celery is not defined
                from achievements.tasks import check_achievement_task
                check_achievement_task.delay(self, user, key, *args, **kwargs)
            else:     
                check_achievement_plain(self, user, key, *args, **kwargs)
        else:
            print "trying to check an achievement for an un-logged user"
                                                     
# create the engine    
engine = AchievementEngine()

        


from django.dispatch import Signal

# An achievement has been registered :
achievement_registered = Signal(providing_args=["achievement_class"])
# A user has unlocked an achievement.
achievement_unlocked = Signal(providing_args=["user", "achievement"])



from distutils.core import setup
#!/usr/bin/env python

try:
    from setuptools import setup, find_packages
    from setuptools.command.test import test
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages
    from setuptools.command.test import test


class mytest(test):
    def run(self, *args, **kwargs):
        from runtests import runtests
        runtests()

setup(
    name = "django-achievements",
    packages = ["achievements", "achievements.migrations", "achievements.templatetags"],
    version = "0.0.7",
    description = "Django application to deal with achievements and rewards unlocked by users",
    author = "Olivier Girardot",
    author_email = "ssaboum@gmail.com",
    url = "https://github.com/ssaboum/django-achievements",
    install_requires = ["importlib>=1.0.2", "django-appconf>=0.4.1"],
    classifiers = [
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from achievements.utils import check_achievement_plain
from celery.task import task  

@task
def check_achievement_task(sender, user, key, *args, **kwargs):
    check_achievement_plain(sender, user, key, *args, **kwargs)